/***************************************************************************
                          irc.c  -  EB-lite's IRC module
                             -------------------
    begin                : Thu Sep 26 2002
    copyright            : (C) 2002 by the Everybuddy team
    email                : www.everybuddy.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


// OK, IRC is a simple protocol. Our challenge is to implement it as simply as possible.

// (One month later, as the number of lines approaches 1000):
// Ha. Ha. Ha.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>

#define GET_LAD(x) ((eb_irc_lad *)(x)->protocol_account_data)
#define GET_GCD(x) ((eb_irc_gcd *)(x)->protocol_data)

#include "services.h"
#include "plugin.h"
#include "plugin_api.h"
#include "gui_comms.h"
#include "elist.h"
#include "dialog.h"
#include "prefs.h"
#include "util.h"
#include "message_parse.h"

typedef struct _eb_irc_lad
{
  int fd;
  char * nick;
  char * server;
  eb_pref_page * config_page;
  eb_contact * last_contact; // If we get an error about messaging, this is who we're talking about
  int input_tag;
} eb_irc_lad;

typedef struct _eb_irc_gcd
{
  char * name;
  char * topic;
} eb_irc_gcd;

int eb_irc_init();
int eb_irc_finish();

int eb_irc_check(void * tag);

void eb_irc_incoming(void * tag, int source, int conditions);
char * eb_irc_read_cmd(int fd, int * line_end);

void eb_irc_incoming_numeric(eb_local_account * ela, int code);

void eb_irc_process_message(eb_local_account * ela, char * from, char * target, char * msg);

eb_plugin_info irc_LTX_plugin_info = {"IRC chat", "Internet Relay Chat service", "(alpha)", "2002-09-26", eb_irc_init, eb_irc_finish};
eb_service_callbacks callbacks;
eb_service irc_service_info = {"IRC", "Meredydd's IRC", &callbacks, SERVICE_CAN_GROUPCHAT, "#8800dd", NULL, NULL, NULL, NULL};
EList * irc_states=NULL;

EList * owned_accounts=NULL;


void eb_irc_set_current_state(eb_local_account * ela, char * state);

void eb_irc_join_group_chat(eb_local_account * ela, char * name);

// eb_irc_init() and eb_irc_finish are at the bottom so I don't have to bother
// prototyping all these functions (lazy little $%& aren't I?)

void eb_irc_setup_local_account(eb_local_account * ela)
{
  eb_irc_lad * lad=(eb_irc_lad *)malloc(sizeof(eb_irc_lad));
  char * tmp;
  char * tmp2;

  lad->nick=strdup(eb_get_value(ela->config_key, "nick"));
  if(lad->nick[0]=='\0')
  {
    char * c;
    free(lad->nick);
    if((c=strstr(ela->handle, "@"))!=NULL)
    {
      *c='\0';
      lad->nick=strdup(ela->handle);
      lad->server=strdup(c+1);
      *c='@'; // change it back...

    } else {
      lad->nick=strdup(ela->handle);
      lad->server=strdup("irc.freenode.net");
    }
    eb_put_value(ela->config_key, "server", lad->server);
    eb_put_value(ela->config_key, "nick", lad->nick);
  } else {
    lad->server=strdup(eb_get_value(ela->config_key, "server"));
  }

  owned_accounts=e_list_append(owned_accounts, ela);
  ela->protocol_account_data=lad;
  lad->fd=-1;
  lad->last_contact=NULL;


  // Set up prefs page
  tmp=(char *)malloc(strlen(lad->nick)+strlen(lad->server)+16);
  tmp2=(char *)malloc(strlen(lad->nick)+strlen(lad->server)+16);
  sprintf(tmp, "%s on %s (IRC)", lad->nick, lad->server);
  sprintf(tmp2, "%s_IRC", ela->handle);
  lad->config_page=eb_make_pref_page(account_prefs, tmp2, tmp);
  free(tmp);
  free(tmp2);
  eb_add_component(lad->config_page, EB_PREF_LABEL, "name_lbl", ela->handle, registry, "params/dump", NULL, NULL);

  eb_add_component(lad->config_page, EB_PREF_TOGGLE, "excl_soa", "Exclude from \"Sign On All\"", ela->config_key, "excl_soa", NULL, NULL);

  eb_add_component(lad->config_page, EB_PREF_STRING, "server", "Server:", ela->config_key, "server", NULL, NULL);
  eb_put_default(ela->config_key, "port", "6667");
  eb_add_component(lad->config_page, EB_PREF_STRING, "port", "Port:", ela->config_key, "port", NULL, NULL);

  eb_add_component(lad->config_page, EB_PREF_STRING, "nick", "Nick (handle):", ela->config_key, "nick", NULL, NULL);
  eb_put_default(ela->config_key, "realname", "Everybuddy user");
  eb_add_component(lad->config_page, EB_PREF_STRING, "realname", "Real name:", ela->config_key, "realname", NULL, NULL);
}

void eb_irc_release_local_account(eb_local_account * ela)
{
  eb_irc_lad * lad=GET_LAD(ela);

  eb_destroy_pref_page(lad->config_page);
  owned_accounts=e_list_remove(owned_accounts, ela);

  free(lad->nick);
  free(lad->server);
  free(lad);
}

void eb_irc_login(eb_local_account * ela)
{
  if(atoi(eb_get_value(ela->config_key, "excl_soa")))
  { return; }

  eb_irc_set_current_state(ela, "Online");
}

void eb_irc_logout(eb_local_account * ela)
{
  eb_irc_lad * lad=GET_LAD(ela);
  EList * n;

  if(ela->connected)
  {
    char msg[]="QUIT :Everybuddy signing off\n";

    eb_input_remove( GET_LAD(ela)->input_tag );
    write(lad->fd, msg, strlen(msg));
    close(lad->fd);
    lad->fd=-1;
  }

  free(ela->status_string);
  ela->status_string=strdup("Offline");

  for(n=ela->buddies; n!=NULL; n=n->next)
  {
    eb_account * acc=(eb_account *)n->data;

    if(acc->status!=EB_ACCOUNT_OFFLINE)
    { eb_buddy_logout(acc); }
  }

  for(n=ela->group_chats; n!=NULL; n=n->next)
  {
    eb_group_chat * chat=(eb_group_chat *)n->data;
    EList * n2;
    EList * queue=NULL;

    for(n2=chat->users; n2!=NULL; n2=n2->next)
    {
      queue=e_list_append(queue, n2->data);
    }

    for(n2=queue; n2!=NULL; n2=n2->next)
    {
      eb_group_chat_left(chat, ((eb_group_chat_user *)n2->data)->handle);
    }

    eb_group_chat_3rdperson(chat, "<font color=red><b><i>Connection to server lost");
  }

  ela->ready=0;
  ela->connected=0;

  eb_local_account_update(ela);
}

char * eb_irc_send_im(eb_account * to, char * message)
{
  eb_irc_lad * lad=GET_LAD(to->buddy_of);
  char * tmp;

  if(!to->buddy_of->ready)
  { eb_show_error("Cannot send messages from an offline account", "IRC error"); free(message); return NULL; }

  if(!strncmp(message, "/me ", 4))
  {
    tmp=(char *)malloc(strlen(to->handle)+strlen(message)+strlen(to->buddy_of->handle)+64);
    sprintf(tmp, "PRIVMSG %s :\001ACTION %s\001\n", to->handle, message+4);
    write(GET_LAD(to->buddy_of)->fd, tmp, strlen(tmp));

    sprintf(tmp, "<font color=#0000ff><b>%s</b></font> %s",
      to->buddy_of->handle, message+4);
    eb_notify_3rdperson(to->contact, tmp);
    free(tmp);
    free(message);

    return NULL;
  } else if(!strncmp(message, "/join #", 7)) {
    eb_irc_join_group_chat(to->buddy_of, message+6);
    free(message);

    return NULL;
  } else {
    tmp=(char *)malloc(strlen(to->handle)+strlen(message)+64);
    sprintf(tmp, "PRIVMSG %s :%s\n", to->handle, message);
    write(lad->fd, tmp, strlen(tmp));
    free(tmp);

    return message;
  }
}

void eb_irc_join_group_chat(eb_local_account * ela, char * name)
{
  char * tmp;

  if(name[0]!='#')
  { eb_show_error("Not a valid channel name - IRC channel names must begin with a hash ('#'), for example \"#everybuddy\"", "IRC error"); return; }

  tmp=(char *)malloc(strlen(name)+32);
  sprintf(tmp, "JOIN %s\n", name);
  write(GET_LAD(ela)->fd, tmp, strlen(tmp));
  free(tmp);
}

void eb_irc_leave_group_chat(eb_group_chat * chat)
{
  char * tmp;

  if(chat->account->ready)
  {
    tmp=(char *)malloc(strlen(GET_GCD(chat)->name)+16);
    sprintf(tmp, "PART %s\n", GET_GCD(chat)->name);
    write(GET_LAD(chat->account)->fd, tmp, strlen(tmp));
    free(tmp);
  }

  eb_destroy_group_chat(chat);
}

char * eb_irc_group_chat_send(eb_group_chat * chat, char * raw)
{
  char * tmp;
  eb_html_item * parsed=eb_html_parse(raw);
  char * msg=eb_html_render_plain(parsed);

  eb_html_destroy(parsed);

  if(strlen(msg)>=500)
  {
    eb_group_chat_3rdperson(chat, "<font color=#ff0000><i><b>Warning:</b> IRC messages must be under 500 characters long. This message has been truncated:</i></font>");
    msg[500]='\0';
  }


  if(!strncmp(msg, "/me ", 4))
  {
    tmp=(char *)malloc(strlen(msg)+strlen(GET_GCD(chat)->name)+strlen(chat->account->handle)+32);
    sprintf(tmp, "PRIVMSG %s :\001ACTION %s\001\n", GET_GCD(chat)->name, msg+4);
    write(GET_LAD(chat->account)->fd, tmp, strlen(tmp));

    sprintf(tmp, "<font color=#8800dd><b>%s</b></font> %s</i>",
      chat->account->handle, msg+4);
    eb_group_chat_3rdperson(chat, tmp);
    free(tmp);
    free(msg);
    free(raw);

    return NULL;
  } else if (!strncmp(msg, "/topic ", 7)) {
    tmp=(char *)malloc(strlen(msg)+strlen(GET_GCD(chat)->name)+32);
    sprintf(tmp, "TOPIC %s :%s\n", GET_GCD(chat)->name, msg+7);
    write(GET_LAD(chat->account)->fd, tmp, strlen(tmp));
    free(tmp);
    free(raw);
    free(msg);

    return NULL;
  } else if(!strncmp(msg, "/join #", 7)) {
    eb_irc_join_group_chat(chat->account, msg+6);
    free(raw);
    free(msg);

    return NULL;
  } else {
    tmp=(char *)malloc(strlen(msg)+strlen(GET_GCD(chat)->name)+32);
    sprintf(tmp, "PRIVMSG %s :%s\n", GET_GCD(chat)->name, msg);
    write(GET_LAD(chat->account)->fd, tmp, strlen(tmp));
    free(tmp);
    free(raw);

    return msg;
  }
}

void eb_irc_set_current_state(eb_local_account * ela, char * state)
{
  if(!strcmp(state, "Online"))
  {
    eb_irc_lad * lad=GET_LAD(ela);
    char * server=eb_get_value(ela->config_key, "server");
    int port=atoi(eb_get_value(ela->config_key, "port"));
    char * nick=eb_get_value(ela->config_key, "nick");
    char * realname=eb_get_value(ela->config_key, "realname");
    char * tmp;

    if(ela->connected) { return; }

    printf("IRC login has begun!\n");

    if(port==0) { eb_show_error("No port specified!", "IRC"); return; }

    printf("Connecting to %s:%d\n", server, port);

    ela->connected=1;
    eb_local_account_update(ela);

    lad->fd=eb_connect_sync_socket(server, port);
    if(lad->fd==-1)
    {
      char * buf=(char *)malloc(strlen(server)+64);
      ela->connected=0;
      eb_local_account_update(ela);
      sprintf(buf, "Connection to %s:%d failed", server, port);
      eb_show_error(buf, "IRC error");
      free(buf);
      return;
    }

    lad->input_tag=eb_input_add(lad->fd, EB_INPUT_READ, eb_irc_incoming, ela);

    tmp=(char *)malloc(strlen(nick)*2+strlen(server)+strlen(realname)+32);
    sprintf(tmp, "NICK %s\nUSER %s 0 * :%s\n", nick, nick, realname);
    write(lad->fd, tmp, strlen(tmp));
    free(tmp);
  }

  if(!strcmp(state, "Offline"))
  {
    if(ela->connected) { eb_irc_logout(ela); }
  }

  if(!strcmp(state, "Away"))
  { eb_show_error("Sorry, the Away state is not yet supported", "IRC error"); }
}

void eb_irc_set_away(eb_local_account * ela, char * short_msg, char * long_msg)
{
}

void eb_irc_unset_away(eb_local_account * ela)
{
}

int eb_irc_init()
{
  printf("Loading IRC module (such as it is...)\n");

  memset(&callbacks, 0, sizeof(callbacks));
  callbacks.setup_local_account=eb_irc_setup_local_account;
  callbacks.release_local_account=eb_irc_release_local_account;
  // no setup or release is required for buddy accounts

  callbacks.login=eb_irc_login;
  callbacks.logout=eb_irc_logout;
  callbacks.send_im=eb_irc_send_im;

  callbacks.set_current_state=eb_irc_set_current_state;
  callbacks.set_away=eb_irc_set_away;
  callbacks.unset_away=eb_irc_unset_away;
  // no add_ or remove_ user calls needed either (three cheers for simplicity!)
  // no concept of block, unblock, or grouping, so no such calls necessary

  callbacks.group_chat_send=eb_irc_group_chat_send;
  callbacks.join_group_chat=eb_irc_join_group_chat;
  callbacks.leave_group_chat=eb_irc_leave_group_chat;

  irc_states=e_list_append(irc_states, strdup("Online"));
  irc_states=e_list_append(irc_states, strdup("Away"));
  irc_states=e_list_append(irc_states, strdup("Offline"));
  irc_service_info.states=irc_states;

  eb_load_service(&irc_service_info);

  eb_set_timer(eb_irc_check, 1, NULL);

  return 0;
}

int eb_irc_finish()
{
  printf("IRC module packing its bags\n");
  return 0;
}

int eb_irc_check(void * tag)
{
  EList * accs;
  for(accs=owned_accounts; accs!=NULL; accs=accs->next)
  {
    eb_local_account * ela=(eb_local_account *)accs->data;
    eb_irc_lad * lad=GET_LAD(ela);

    if(!ela->ready) { continue; }

    if(ela->buddies!=NULL)
    {
      eb_account * acc=(eb_account *)ela->buddies->data;
      char * whocmd=(char *)malloc(strlen(acc->handle)+16);

      sprintf(whocmd, "WHO %s\n", acc->handle);
      write(lad->fd, whocmd, strlen(whocmd));
      free(whocmd);

      // the response to this first WHO will result in sending the next WHO, and so on
    }
  }

  return 30;
}

eb_group_chat * eb_irc_get_chat(eb_local_account * acc, char * name)
{
  EList * n;

  for(n=group_chats; n!=NULL; n=n->next)
  {
    eb_group_chat * chat=(eb_group_chat *)n->data;
    if(chat->account!=acc) { continue; }

    if(!strcasecmp(GET_GCD(chat)->name, name)) { return chat; }
  }

  return NULL;
}

void eb_irc_trim_nick(char * s)
{
  int pos=0;

  while(1)
  {
    if(s[pos]=='!') { s[pos]='\0'; return; }
    if(s[pos]=='\0') { return; }
    pos++;
  }
}

void eb_irc_incoming(void * tag, int fd, int conditions)
{
  int line_end=0;
  char * parm;
  char * source=NULL;
  eb_local_account * ela=(eb_local_account *)tag;

  parm=eb_irc_read_cmd(fd, &line_end);

  if(parm==NULL)
  {
    eb_irc_logout(ela);
    return;
  }

  if(parm[0]==':')
  {
    source=parm;
    if(line_end)
    {
      printf("IRC: WTF?!\n%s\n", parm);
      free(parm);
      return;
    }

    parm=eb_irc_read_cmd(fd, &line_end);
  }


  if(isdigit(parm[0]))
  {
    eb_irc_incoming_numeric(ela, atoi(parm));
    free(parm);
  } else if(!strcmp(parm, "JOIN")) {
    char * channel;
    char * buf;
    eb_group_chat * chat;
    eb_account * buddy;

    channel=eb_irc_read_cmd(fd, &line_end);
    chat=eb_irc_get_chat(ela, channel+1);

    eb_irc_trim_nick(source);

    if(chat==NULL)
    {
      chat=eb_create_group_chat(ela);
      chat->protocol_data=(eb_irc_gcd *)malloc(sizeof(eb_irc_gcd));
      chat->title=strdup(channel+1);
      GET_GCD(chat)->name=strdup(channel+1);
      GET_GCD(chat)->topic=strdup("No topic set");
    }

    if(chat->announced)
    {
      buf=(char *)malloc(strlen(source)+64);
      sprintf(buf, "<font color=#666666><i><b>%s</b> has joined the channel", source+1);
      eb_group_chat_3rdperson(chat, buf);
      free(buf);
      eb_group_chat_joined(chat, source+1);
    }

    // Now check whether the recent joiner is on our buddy list, and sign them in
    // - if they've joined a channel with us, they're online
    buddy=eb_get_account(ela->handle, ela->service_name, source+1);
    if(buddy!=NULL)
    {
      if(buddy->status==EB_ACCOUNT_OFFLINE)
      {
        free(buddy->status_string);
        buddy->status_string=strdup("");
        eb_buddy_login(buddy);
        eb_buddy_update_status(buddy);
      }
    }

    free(channel);
    free(source);
    free(parm);
  } else if(!strcmp(parm, "PART")) {
    char * channel;
    eb_group_chat * chat;

    channel=eb_irc_read_cmd(fd, &line_end);
    chat=eb_irc_get_chat(ela, channel);
    free(channel);

    eb_irc_trim_nick(source);

    if(chat!=NULL)
    {
      char * buf=(char *)malloc(strlen(source)+64);
      sprintf(buf, "<font color=#666666><i><b>%s</b> has left the channel", source+1);
      eb_group_chat_3rdperson(chat, buf);
      free(buf);
      eb_group_chat_left(chat, source+1);
    }

    free(source);
    free(parm);
  } else if(!strcmp(parm, "NICK")) {
    eb_account * buddy;
    EList * n;
    char * new_nick=eb_irc_read_cmd(fd, &line_end);

    if(source==NULL)
    { printf("IRC: nick change from nowhere. Odd, that...\n"); free(new_nick); free(parm); return; }

    eb_irc_trim_nick(source);

    for(n=group_chats; n!=NULL; n=n->next)
    {
      eb_group_chat * chat=(eb_group_chat *)n->data;
      EList * n2;

      for(n2=chat->users; n2!=NULL; n2=n2->next)
      {
        eb_group_chat_user * user=(eb_group_chat_user *)n2->data;

        if(!strcmp(user->handle, source+1))
        {
          char * buf=(char *)malloc(strlen(source)+64);
          sprintf(buf, "<font color=#666666><i><b>%s</b> is now known as <b>%s</b>", source+1, new_nick+1);
          eb_group_chat_3rdperson(chat, buf);
          free(buf);

          eb_group_chat_left(chat, source+1);
          eb_group_chat_joined(chat, new_nick+1);
          break;
        }
      }
    }

    buddy=eb_get_account(ela->handle, "IRC", source+1);
    if(buddy!=NULL && buddy->status!=EB_ACCOUNT_OFFLINE)
    { buddy->status=EB_ACCOUNT_OFFLINE; eb_buddy_logout(buddy); }

    buddy=eb_get_account(ela->handle, "IRC", new_nick+1);
    if(buddy!=NULL && buddy->status==EB_ACCOUNT_OFFLINE)
    { 
      free(buddy->status_string);
      buddy->status_string=strdup("");
      eb_buddy_login(buddy);
      eb_buddy_update_status(buddy);
    }

    free(new_nick);
    free(source);
    free(parm);
  } else if(!strcmp(parm, "TOPIC")) {
    char * channel=eb_irc_read_cmd(fd, &line_end);
    char * topic=eb_irc_read_cmd(fd, &line_end);
    eb_group_chat * chat=eb_irc_get_chat(ela, channel);
    char * tmp;

    if(chat==NULL) { free(source); free(parm); return; }

    eb_irc_trim_nick(source);

    tmp=(char *)malloc(strlen(source)+strlen(channel)+strlen(topic)+128);
    sprintf(tmp, "<i>%s has changed the topic of %s to:<br><font color=#4444ff>%s</font></i>",
      source+1, channel, topic+1);
    eb_group_chat_3rdperson(chat, tmp);
    free(tmp);

    free(GET_GCD(chat)->topic);
    GET_GCD(chat)->topic=strdup(topic+1);
    free(topic);

    free(source);
    free(channel);
    free(parm);
  } else if(!strcmp(parm, "QUIT")) {
    eb_account * buddy;
    EList * n;

    if(source==NULL)
    { printf("IRC: quit from nowhere. Odd, that...\n"); free(parm); return; }

    eb_irc_trim_nick(source);

    for(n=group_chats; n!=NULL; n=n->next)
    {
      eb_group_chat * chat=(eb_group_chat *)n->data;
      EList * n2;

      for(n2=chat->users; n2!=NULL; n2=n2->next)
      {
        eb_group_chat_user * user=(eb_group_chat_user *)n2->data;

        if(!strcmp(user->handle, source+1))
        {
          // cmb: can't figure out why, but second works, first doesn't!
          //      (doesn't as in tells me 'testquit' for 'cmb|testquit')
          //eb_group_chat_left(chat, user->handle);
          char * buf=(char *)malloc(strlen(source)+64);
          sprintf(buf, "<font color=#666666><i><b>%s</b> has quit IRC", source+1);
          eb_group_chat_3rdperson(chat, buf);
          free(buf);

          eb_group_chat_left(chat, source+1);
          break;
        }
      }
    }

    buddy=eb_get_account(ela->handle, "IRC", source+1);
    if(buddy!=NULL && buddy->status!=EB_ACCOUNT_OFFLINE)
    { buddy->status=EB_ACCOUNT_OFFLINE; eb_buddy_logout(buddy); }

    free(source);
    free(parm);

    while(!line_end)
    { free(eb_irc_read_cmd(fd, &line_end)); }
  } else if(!strcmp(parm, "PRIVMSG")) {
    char * msg;
    char * target;

    if(source==NULL) { printf("IRC: PRIVMSG without source!\n"); free(parm); return; }

    eb_irc_trim_nick(source);
    target=eb_irc_read_cmd(fd, &line_end);
    msg=eb_irc_read_cmd(fd, &line_end);

    eb_irc_process_message(ela, source+1, target, msg+1);

    free(msg);
    free(target);
    free(source);
    free(parm);
  } else if(!strcmp(parm, "NOTICE")) {
    char * msg;
    char * target;

    if(source==NULL) { printf("IRC: NOTICE without source!\n"); free(parm); return; }

    // TODO: not containing an at, i.e. server notice
    //       (eventually -> server console, a hidden groupchat?)
    if (!strchr(source, '@')) { free(source); free(parm); return; }

    eb_irc_trim_nick(source);
    target=eb_irc_read_cmd(fd, &line_end);
    msg=eb_irc_read_cmd(fd, &line_end);

    // TODO: indicate at all that it was a NOTICE?
    eb_irc_process_message(ela, source+1, target, msg+1);

    free(msg);
    free(target);
    free(source);
    free(parm);
  } else if(!strcmp(parm, "PING")) {
    char * tmp;

    free(parm);
    parm=eb_irc_read_cmd(fd, &line_end);

    tmp=(char *)malloc(strlen(parm)+16);
    sprintf(tmp, "PONG %s\n", parm);
    write(GET_LAD(ela)->fd, tmp, strlen(tmp));
    free(tmp);
    free(parm);

    if(source!=NULL) { free(source); }

  } else if(!strcmp(parm, "ERROR")) {
    free(parm);
    parm=eb_irc_read_cmd(fd, &line_end);
    eb_show_error(parm+1, "IRC error");
    free(parm);
    if(source!=NULL) { free(source); }
    eb_irc_logout(ela);
    return;
  } else {
/*
    printf("Unknown IRC line, take a look:\n");
    while(1)
    {
      printf("%s ", parm);
      free(parm);

      if(line_end) { break; }

      parm=eb_irc_read_cmd(fd, &line_end);
    }
*/

    free(parm);
    if(source!=NULL) { free(source); }
  }
}

void eb_irc_incoming_numeric(eb_local_account * ela, int code)
{
  eb_irc_lad * lad=GET_LAD(ela);
  int line_end=0;

  // OK this is a hack for buddy checks, but it works, and it conforms to the RFC
  static int last_is_online=0;

  if(code==401) // "no such user/channel"
  {
    char * buddy;
    eb_account * acc;

    free(eb_irc_read_cmd(lad->fd, &line_end)); // my nickname, I know that

    buddy=eb_irc_read_cmd(lad->fd, &line_end);
    acc=eb_get_account(ela->handle, "IRC", buddy);
    free(buddy);

    if(acc==NULL) { return; }

    eb_notify_3rdperson(acc->contact, "<font color=red><i><b>Warning:</b> Your last message failed, because the user has logged off from IRC.</i></font>");
    eb_buddy_logout(acc);

    return;
  }

  if(code>=400 && code<500) // it's an error
  {
    char * msg=strdup("IRC server error: ");
    char * parm;
    char * tmp;

    while(!line_end)
    {
      parm=eb_irc_read_cmd(GET_LAD(ela)->fd, &line_end);
      if(!strcmp(parm, "*")) { continue; }
      tmp=(char *)malloc(strlen(msg)+strlen(parm)+2);
      strcpy(tmp, msg);
      strcat(tmp, " ");
      if(parm[0]==':') { strcat(tmp, parm+1); } else { strcat(tmp, parm); }
      free(msg);
      free(parm);
      msg=tmp;
    }

    eb_show_error(msg, "IRC error");
    free(msg);
    return;
  }

  if(code==1)
  {
    EList * n;

    ela->ready=1;
    free(ela->status_string);
    ela->status_string=strdup("Online");
    eb_local_account_update(ela);

    for(n=ela->group_chats; n!=NULL; n=n->next)
    {
      eb_group_chat * chat=(eb_group_chat *)n->data;

      eb_irc_join_group_chat(ela, chat->title);
    }
  }

  if(code==352)
  {
    while(!line_end)
    { free(eb_irc_read_cmd(lad->fd, &line_end)); }

    last_is_online=1;
  }

  if(code==315)
  {
    char * buddy_nick;
    eb_account * acc=NULL;
    EList * n;

    free(eb_irc_read_cmd(lad->fd, &line_end)); // my nick, I already know it

    buddy_nick=eb_irc_read_cmd(lad->fd, &line_end);

    while(!line_end)
    { free(eb_irc_read_cmd(lad->fd, &line_end)); }

    for(n=ela->buddies; n!=NULL; n=n->next)
    {
      acc=(eb_account *)n->data;
      if(!strcmp(acc->handle, buddy_nick)) { break; }
      acc=NULL;
    }

    if(acc==NULL)
    {
      printf("Strange - I *swear* I used to know %s...\n", buddy_nick);
      free(buddy_nick);
      last_is_online=0;
      return;
    }

    // OK, if we know this is a valid account, n is still pointing to the next node

    if(n->next!=NULL)
    {
      eb_account * next_acc=(eb_account *)(n->next->data);
      char * whocmd=(char *)malloc(strlen(next_acc->handle)+16);

      sprintf(whocmd, "WHO %s\n", next_acc->handle);
      write(lad->fd, whocmd, strlen(whocmd));
      free(whocmd);
    }

    if(last_is_online)
    {
      if(acc->status==EB_ACCOUNT_OFFLINE)
      {
        free(acc->status_string);
        acc->status_string=strdup("");
        eb_buddy_login(acc);
        eb_buddy_update_status(acc);
      }
    } else {
      if(acc->status!=EB_ACCOUNT_OFFLINE)
      {
        free(acc->status_string);
        acc->status_string=strdup("(Offline)");
        eb_buddy_logout(acc);
        eb_buddy_update_status(acc);
      }
    }

    last_is_online=0;
  }

  if(code==332) // topic
  {
    char * channel;
    char * topic;
    eb_group_chat * chat;

    free(eb_irc_read_cmd(lad->fd, &line_end)); // drop nick

    channel=eb_irc_read_cmd(lad->fd, &line_end);
    chat=eb_irc_get_chat(ela, channel);

    if(chat==NULL) { return; }

    free(GET_GCD(chat)->topic);
    topic=eb_irc_read_cmd(lad->fd, &line_end);
    GET_GCD(chat)->topic=strdup(topic+1);
    free(topic);
  }

  if(code==353) // RPM_NAMEREPLY
  {
    char * channel;
    eb_group_chat * chat;
    char * s;
    char * last_user;
    char * user_list;

    free(eb_irc_read_cmd(lad->fd, &line_end)); // drop username
    free(eb_irc_read_cmd(lad->fd, &line_end)); // drop channel type

    channel=eb_irc_read_cmd(lad->fd, &line_end);
    chat=eb_irc_get_chat(ela, channel);
    free(channel);

    if(chat==NULL)
    { printf("NAMES on unknown channel, ignoring\n"); return; }

    user_list=eb_irc_read_cmd(lad->fd, &line_end);
    s=last_user=user_list+1;

    while(1)
    {
      if(*s==' ' || *s=='\0')
      {
        int end=(*s=='\0');

        *s='\0';

        if(last_user[0]=='@')
        {
          eb_group_chat_joined(chat, last_user+1);
          // TODO: flag them as an op
        } else {
          eb_group_chat_joined(chat, last_user);
        }
        last_user=s+1;

        if(end || last_user[0]=='\0') { break; }
      }

      s++;
    }

    free(user_list);
  }

  if(code==366)
  {
    char * tmp;
    char * channel;
    eb_group_chat * chat;

    free(eb_irc_read_cmd(lad->fd, &line_end)); // drop username

    channel=eb_irc_read_cmd(lad->fd, &line_end);
    chat=eb_irc_get_chat(ela, channel);
    free(channel);

    if(chat==NULL) { return; }

    eb_announce_group_chat(chat);

    tmp=(char *)malloc(strlen(GET_GCD(chat)->topic)+64);
    sprintf(tmp, "<i>Topic for today:</i>\n<font color=#4444ff>%s</font>\n", GET_GCD(chat)->topic);
    eb_group_chat_3rdperson(chat, tmp);
    free(tmp);
  }
}

void eb_irc_process_message(eb_local_account * ela, char * from, char * target, char * msg)
{
  eb_group_chat * chat=eb_irc_get_chat(ela, target);
  eb_account * acc=NULL;
  int start=0;
  int pos=0;

  if(chat==NULL)
  {
    acc=eb_get_account(ela->handle, "IRC", from);
    if(acc==NULL)
    {
      eb_group * group=eb_get_group("Unknown");
      eb_contact * cont;

      if(group==NULL)
      { group=eb_add_group("Unknown"); }

      cont=eb_get_contact("Unknown", from);
      if(cont==NULL)
      { cont=eb_add_contact(group, from); }

      acc=(eb_account *)malloc(sizeof(eb_account));
      acc->handle=strdup(from);
      acc->status_string=strdup("");
      acc->status=EB_ACCOUNT_ONLINE;
      eb_add_account(cont, ela, acc);
      eb_buddy_login(acc);
    } else if(acc->status==EB_ACCOUNT_OFFLINE) {
      // they msged us, they're online
      free(acc->status_string);
      acc->status_string=strdup("");
      eb_buddy_login(acc);
      eb_buddy_update_status(acc);
    }
  }

  // OK, now let's get parsing!

  while(msg[pos]!='\0')
  {
    start=pos;

    while(1)
    {
      pos++;
      if(msg[pos]=='\0' || msg[pos]==1) { break; }
    }

    // OK, now start is at the beginning of this segment, and pos is at the end.
    if(msg[start]==1) // CTCP
    {
      if(pos-start>7 && !strncmp(msg+start+1, "ACTION ", 7))
      {
        char * buf=(char *)malloc(512);
        char html_buf[1024]; // fixed size justified because of IRC line length limits

        strncpy(buf, msg+start+8, pos-start-8);
        buf[pos-start-8]='\0';

        if(msg[pos]==1) { pos++; }

        if(chat!=NULL)
        {
          buf=eb_filter_string(EB_GROUP_FILTER_IN, buf, chat);
          if(buf==NULL) { continue; }
          sprintf(html_buf, "<font color=#8800dd><b>%s</b></font> %s", from, buf);
          eb_group_chat_3rdperson(chat, html_buf);
        } else {
          buf=eb_filter_string(EB_FILTER_IN, buf, acc);
          if(buf==NULL) { continue; }
          sprintf(html_buf, "<font color=#8800dd><b>%s</b></font> %s",
            acc->contact->name, buf);
          eb_notify_3rdperson(acc->contact, html_buf);
        }

        free(buf);
        continue;
      }

      // If we get here, it's an unrecognised type of CTCP. Complain. Loudly.

      if(chat)
      {
        eb_group_chat_3rdperson(chat, "<font color=red><b>** Received unknown CTCP command, ignoring it.</b></font><br>");
      } else {
        eb_notify_3rdperson(acc->contact, "<font color=red><b>** Received unknown CTCP command, ignoring it.</b></font><br>");
      }

    } else { // normal message
      char * buf=(char *)malloc(512);

      strncpy(buf, msg+start, pos-start);
      buf[pos-start]='\0';


      if(chat!=NULL)
      {
        buf=eb_filter_string(EB_GROUP_FILTER_IN, buf, chat);
        if(buf==NULL) { continue; }
        eb_group_chat_message(chat, from, buf);
      } else {
        buf=eb_filter_string(EB_FILTER_IN, buf, acc);
        if(buf==NULL) { continue; }
        eb_incoming_message(acc, buf);
      }

      free(buf);
    }
  }
}


char * eb_irc_read_cmd(int fd, int * line_end)
{
  static char buffer[4096];
  static int bol=1; // beginning of line
  int pos=0;
  int trailer=0;

  *line_end=0;
  buffer[0]='\0';

  while(1)
  {
    char c;

    if(read(fd, &c, 1)<1)
    {
      printf("IRC - Connection closing\n");
      return NULL;
    }

    switch(c)
    {
      case('\r') :
      case('\n') : {
        *line_end=1;
        buffer[pos]='\0';
        bol=1;
        return strdup(buffer);
      }

      case(' ') : {
        if(trailer)
        {
          buffer[pos++]=' ';
          if(pos>=4095)
          { printf("IRC: Dropping characters!\n"); pos--; }
          break;
        }

        if(pos==0) { continue; }

        buffer[pos]='\0';
        bol=0;
        return strdup(buffer);
      }

      default : {
        if(pos==0 && c==':' && bol==0) { trailer=1; }
        buffer[pos++]=c;
        if(pos>=4095)
        { printf("IRC: Dropping characters!\n"); pos--; }
      }
    }
  }
}
