/***************************************************************************
                      message_parse.c - Filter messages
                             -------------------
                     (C) 2002 by the Everybuddy team
                            www.everybuddy.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "globals.h"
#include "gui_comms.h"
#include "message_parse.h"

// held messages
EList * eb_held_messages=NULL;
int eb_holding_messages=1;

// filter chains
EList * eb_filters[6]={NULL, NULL, NULL, NULL, NULL, NULL};

char * eb_prepare_away_msg(void);

void eb_incoming_message(eb_account * remote, char * message)
{
  time_t now=time(NULL);

  if(remote->contact->ignore!=NULL)
  {
    if(remote->contact->next_away_msg>now) { return; }

    if(remote->buddy_of->service->sc->send_im==NULL) { return; }

    if(remote->contact->ignore[0]!='\0')
    {
      char header[]="This is an automated message. The person you attempted to contact is ignoring you and did not receive your message.\nExplanation: ";
      char * msg=(char *)malloc(strlen(header)+strlen(remote->contact->ignore)+1);
      strcpy(msg, header);
      strcat(msg, remote->contact->ignore);
      remote->buddy_of->service->sc->send_im(remote, msg);
      // do NOT acknowledge this with eb_gui_sent_message(), otherwise we lose the whole point...
      remote->contact->next_away_msg=now+60;
      free(msg);
    }

    return;
  }

  if(eb_holding_messages)
  {
    eb_held_message * msg=(eb_held_message *)malloc(sizeof(eb_held_message));
    msg->buddy=remote;
    msg->contact=NULL;
    msg->message=strdup(message);
    msg->when=time(NULL);
    msg->direction=EB_MSG_RECV;
    remote->locked++;

    eb_held_messages=e_list_append(eb_held_messages, msg);
  } else {
    eb_gui_got_message(remote, message);
  }

  if(away_short!=NULL)
  {
    if(remote->contact->next_away_msg<now)
    {
      char * away_sent;

      remote->contact->next_away_msg=now+300;
      if(remote->buddy_of->service->sc->send_im==NULL) { return; }

      away_sent=remote->buddy_of->service->sc->send_im(remote, eb_prepare_away_msg());
      if(away_sent==NULL) { return; }

      if(eb_holding_messages)
      {
        eb_held_message * msg=(eb_held_message *)malloc(sizeof(eb_held_message));
        msg->buddy=remote;
        msg->contact=NULL;
        msg->message=strdup(away_sent);
        msg->when=time(NULL);
        msg->direction=EB_MSG_SEND;
        remote->locked++;

        eb_held_messages=e_list_append(eb_held_messages, msg);
      } else {
        eb_gui_sent_message(remote, away_sent);
      }

      free(away_sent);
    }
  }
}

char * eb_prepare_away_msg(void)
{
  int len=0, a, b;
  char * msg;
  
  printf("Preparing a message\n");
  
  for(a=0; away_msg[a]!='\0'; a++)
  {
    if(away_msg[a]=='%')
    {
      switch(away_msg[a+1])
      {
        case('t') : { len+=16; break; }
      }
    }
    len++;
  }
  
  msg=(char *)malloc(len+1);
  
  for(a=0, b=0; away_msg[a]!='\0'; a++)
  {
    if(away_msg[a]=='%')
    {
      switch(away_msg[a+1])
      {
        case('t') : {
	  time_t now=time(NULL);
	  struct tm * tv=localtime(&now);
	  
	  b+=sprintf(msg+b, "%d:%02d", tv->tm_hour, tv->tm_min);
	  a++;
	  continue;
	}
	
	case('%') : { a++; }
      }
    }
    
    msg[b++]=away_msg[a];
  }
  msg[b]='\0';
  
  return msg;
}

void eb_notify_3rdperson(eb_contact * contact, char * message)
{
  if(contact->ignore!=NULL) { return; }

  if(eb_holding_messages)
  {
    eb_held_message * msg=(eb_held_message *)malloc(sizeof(eb_held_message));
    msg->buddy=NULL;
    msg->contact=contact;
    msg->message=strdup(message);
    msg->when=time(NULL);
    msg->direction=EB_MSG_3RDPERSON;
    contact->locked++;

    eb_held_messages=e_list_append(eb_held_messages, msg);
  } else {
    eb_gui_notify_3rdperson(contact, message);
  }
}


// Filter chain manipulation
// call a filter chain on a string - this string may be free()ed by those functions
char * eb_filter_string(int chain_id, char * string, void * context)
{
  char * s=string;
  EList * n;

  for(n=eb_filters[chain_id]; n!=NULL; n=n->next)
  {
    eb_filter * tf=(eb_filter *)n->data;

    s=tf->func(s, context);
    if(s==NULL) { return NULL; }
  }

  return s;
}

// Add a filter to chain
void eb_add_filter(int chain_id, eb_filter_func func, int priority)
{
  eb_filter * filter=(eb_filter *)malloc(sizeof(eb_filter));
  EList * n;

  filter->func=func;
  filter->priority=priority;

  for(n=eb_filters[chain_id]; n!=NULL; n=n->next)
  {
    eb_filter * tf=(eb_filter *)n->data;

    if(tf->priority>priority)
    { eb_filters[chain_id]=e_list_insert(eb_filters[chain_id], n, filter); return; }
  }

  // if we got this far, this is the highest-priority link in the chain (ie executed last)

  eb_filters[chain_id]=e_list_append(eb_filters[chain_id], filter);
}

// Remove a filter from a chain
void eb_del_filter(int chain_id, eb_filter_func func)
{
  EList * n;

  for(n=eb_filters[chain_id]; n!=NULL; n=n->next)
  {
    eb_filter * tf=(eb_filter *)n->data;

    if(tf->func==func)
    {
      eb_filters[chain_id]=e_list_remove_link(eb_filters[chain_id], n);
      e_list_free_1(n);
      free(tf);
      return;
    }
  }

  eb_debug(DBG_CORE, "Cannot find filter to delete!\n");
}

