#ifndef __MESSAGE_PARSE_H__
#define __MESSAGE_PARSE_H__

#include <time.h>

#include "contactlist.h"
#include "html.h"

#ifdef __cplusplus
extern "C" {
#endif

#define EB_MSG_SEND      0
#define EB_MSG_RECV      1
#define EB_MSG_3RDPERSON 2

typedef struct _eb_held_message {
  eb_account * buddy;
  eb_contact * contact;
  char * message;
  time_t when;
  int direction;
} eb_held_message;

extern EList * eb_held_messages;

extern int eb_holding_messages;

typedef char * (*eb_filter_func) (char * string, void * context);
// the void * can be an eb_account (in one-to-one chat), or
// an eb_group_chat (in groupchat send), or an eb_group_chat_user
// (in groupchat receive). Cast it right, my friend...

typedef struct _eb_filter {
  eb_filter_func func;
  int priority; // UNIX-style: -20 to 20, low means called first. 0 is a sensible default
} eb_filter;

// filter chains

#define EB_FILTER_IN		0
#define EB_PREFILTER_OUT	1
#define EB_POSTFILTER_OUT	2

#define EB_GROUP_FILTER_IN	3
#define EB_GROUP_PREFILTER_OUT	4
#define EB_GROUP_POSTFILTER_OUT	5

extern EList * eb_filters[];

// Incoming IM - this will appear in the normal fashion, with the relevant user/contact's
// name. If this is called on an ignored contact, a bounce message may be sent
void eb_incoming_message(eb_account * remote, char * message);

// 3rd person commentary to appear in the chat window with no other frills (apart from maybe
// a timestamp, at the UI's discretion). If this is called on an ignored contact,
// it will be silently dropped with no bounce
void eb_notify_3rdperson(eb_contact * contact, char * message);


// Filter chains

// call a filter chain on a string - this string may be free()ed by those functions
char * eb_filter_string(int chain_id, char * string, void * context);

// Add a filter to chain
void eb_add_filter(int chain_id, eb_filter_func func, int priority);

// Remove a filter from a chain
void eb_del_filter(int chain_id, eb_filter_func func);

#ifdef __cplusplus
}
#endif

#endif // __MESSAGE_PARSE_H__
