/***************************************************************************
                          proxy.c - Connection wrappers
                             -------------------
                     (C) 2002 by the Everybuddy team
                            www.everybuddy.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>

#include "plugin_api.h"
#include "dialog.h"

typedef struct _eb_rawtcp_data {
  void (*cb)(int s, void * data);
  void * data;
  int itag;
} eb_rawtcp_data;

typedef struct _eb_socks5_data {
  char * host;
  int port;
  void (*cb)(int s, void * data);
  void * data;
  int nextstage;
  int itag;
} eb_socks5_data;

typedef struct _eb_http_data {
  void (*cb)(int s, void * data);
  void * data;
  char * url;
  char * host;
  int port;
  int method;
} eb_http_data;

void eb_rawtcp_finish(void * cdata, int s, int conditions)
{
  eb_rawtcp_data * data=(eb_rawtcp_data *)cdata;
  int err;
  int val=0, len=4;

  if((err=getsockopt(s, SOL_SOCKET, SO_ERROR, &val, &len))<0 || val!=0)
  {
    close(s);
    data->cb(-1, data->data);
  } else {
    fcntl(s, F_SETFL, 0);
    data->cb(s, data->data);
  }

  eb_input_remove(data->itag);

  free(data);
}

void eb_rawtcp_connect(char * hostname, int port, void (*cb)(int s, void * data), void * cdata)
{
  eb_rawtcp_data * edata;
  struct sockaddr_in sa;
  struct hostent     *hp;
  int s;

  // This should be asynchronous. I could *murder* the fool who reckoned that
  // making gethostbyname() synchronous-only was a good idea...

  if ((hp= gethostbyname(hostname)) == NULL) { /* do we know the host's */
    errno= ECONNREFUSED;                       /* address? */
    cb(-1, cdata);                                /* no */
    return;
  }

  memset(&sa,0,sizeof(sa));
  memcpy((char *)&sa.sin_addr,hp->h_addr,hp->h_length);     /* set address */
  sa.sin_family= hp->h_addrtype;
  sa.sin_port= htons((u_short)port);

  if ((s= socket(hp->h_addrtype,SOCK_STREAM,0)) < 0)     /* get socket */
  {
    cb(-1, cdata);
    return;
  }

  //fcntl(s, F_SETFL, O_NONBLOCK);

  if (connect(s,(struct sockaddr *)&sa,sizeof sa) >= 0) { /* connect */
    // wow, instant response! Cool beans...
    fcntl(s, F_SETFL, 0);
    cb(s, cdata);
    return;
  }
  // Otherwise, we're just gonna have to wait for it
  if(errno!=EINPROGRESS)
  {
    close(s);
    cb(-1, cdata);
    return;
  }

  edata=(eb_rawtcp_data *)malloc(sizeof(eb_rawtcp_data));
  edata->data=cdata;
  edata->cb=cb;
  edata->itag=eb_input_add(s, EB_INPUT_READ | EB_INPUT_WRITE |  EB_INPUT_EXCEPTION, eb_rawtcp_finish, edata);
}

/*****************************************************************************/

void eb_socks5_incoming(void * vdata, int s, int conditions)
{
  eb_socks5_data * data=(eb_socks5_data *)vdata;
  char incoming[16];
  unsigned char * buf;
  int a;

  if(read(s, incoming, 2)<2 || incoming[1]!=0)
  {
    eb_input_remove(data->itag);
    close(s);
    data->cb(-1, data->data);
    free(data->host);
    free(data);
    return;
  }

  if(data->nextstage)
  {
    read(s, incoming, 2);
    switch(incoming[1]) {
      case(1) : { // IPV4
        read(s, incoming, 6);
        break;
      }

      case(3) : { // DNS
        int a;
        unsigned char n;

        read(s, &n, 1);
        for(a=0; a<n+2; a++)
        {
          read(s, incoming, 1);
        }
        break;
      }

      case(4) : {
        read(s, incoming, 16);
        break;
      }

      default : {
        eb_show_error("The SOCKS proxy gave an unknown address type", "Proxy error");
        eb_input_remove(data->itag);
        close(s);
        data->cb(-1, data->data);
        free(data->host);
        free(data);
        return;
      }
    }
    // OK folks, that's it! This TCP connection now belongs to the server and the routine which asked
    // for it. We just need to get out of the way.

    eb_input_remove(data->itag);
    data->cb(s, data->data);
    free(data->host);
    free(data);
    return;
  }

  buf=(char *)malloc(strlen(data->host)+7);
  buf[0]=5; // version
  buf[1]=1; // "CONNECT" command
  buf[2]=0; // reserved
  buf[3]=3; // this is a DNS addy
  buf[4]=strlen(data->host);
  for(a=0; data->host[a]!=0; a++)
  {
    buf[5+a]=data->host[a];
  }
  buf[5+a++]=(data->port)/256;
  buf[5+a++]=(data->port)%256;

  write(s, buf, a+5);

  free(buf);
  data->nextstage=1;
}

void eb_socks5_connect_cb(int s, void * vdata)
{
  eb_socks5_data * data=(eb_socks5_data *)vdata;
  char send[]={5, 1, 0}; // version 5, 1 authentication method supported, unauthenticated

  if(s==-1)
  {
    data->cb(-1, data->data);
    free(data->host);
    free(data);
    return;
  }

  write(s, send, 3);
  data->nextstage=0;

  data->itag=eb_input_add(s, EB_INPUT_READ | EB_INPUT_EXCEPTION, eb_socks5_incoming, data);
}

void eb_socks5_connect(char * hostname, int port, void (*cb)(int s, void * data), void * cdata)
{
  eb_socks5_data * data=(eb_socks5_data *)malloc(sizeof(eb_socks5_data));
  data->host=strdup(hostname);
  data->port=port;
  data->cb=cb;
  data->data=cdata;

  eb_rawtcp_connect(eb_get_value(registry, "config/proxy/socks5/host"), atoi(eb_get_value(registry, "config/proxy/socks5/port")), eb_socks5_connect_cb, data);
}


/*****************************************************************************/

void eb_sync_socket_cb(int s, void * data)
{
  *((int *)data)=s;
}

void eb_connect_socket(char * hostname, int port, void (*cb)(int s, void * data), void * cdata)
{
  eb_rawtcp_connect(hostname, port, cb, cdata);
//  eb_socks5_connect(hostname, port, cb, cdata);
}


int eb_connect_sync_socket(char * hostname, int port)
{
  int done=-2;

  eb_connect_socket(hostname, port, eb_sync_socket_cb, &done);

  while(done==-2)
  { eb_polling_loop(); } // my soul's going to rot in hell for this...be CAREFUL, people!

  return done;
}

/*****************************************************************************/

void eb_http_cb(int s, void * data)
{
  eb_http_data * hdata=(eb_http_data *)data;
  char * request;
  int len;

  if(s>=0)
  {
    request=malloc(64+strlen(hdata->url)+strlen(hdata->host));
    len=sprintf(request, "%s %s HTTP/1.1\r\nHost: %s:%d\r\n",
      (hdata->method==EB_HTTP_GET)?("GET"):("POST"), hdata->url, hdata->host, hdata->port);

    write(s, request, len);
  }

  hdata->cb(s, hdata->data);

  free(hdata->host);
  free(hdata->url);
  free(hdata);
}

void eb_http_request(char * host, int port, char * url, int method, void (*cb)(int s, void * data), void * data)
{
  eb_http_data * hdata=(eb_http_data *)malloc(sizeof(eb_http_data));

  hdata->cb=cb;
  hdata->data=data;
  hdata->url=strdup(url);
  hdata->method=method;
  hdata->host=strdup(host);
  hdata->port=port;

  eb_connect_socket(host, port, eb_http_cb, hdata);
}
