/*
 *   PCM time-domain equalizer
 *
 *   Copyright (C) 2002  Felipe Rivera <liebremx at users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *   $Id: eq.c,v 1.3 2003/09/10 19:05:02 liebremx Exp $
 */

#include "eq.h"

float gain[EQ_MAX_BANDS][EQ_CHANNELS] __attribute__((aligned));
float preamp[EQ_CHANNELS] __attribute__((aligned));
static gboolean on = FALSE;
static void init(void);
static void cleanup(void);
static gint raisevol(gpointer data);
static int volume_diff;

EffectPlugin eq_ep =
{
	NULL,
	NULL,
	NULL, /* Description */
	init,
	cleanup,
	about,
	eq_configuration_window,
	mod_samples,
	NULL
};


EffectPlugin *get_eplugin_info(void)
{
	eq_ep.description = g_strdup_printf("EQ Plugin %s", VERSION);
	return &eq_ep;
}

static void init(void)
{
	/* Init the GUI first to load the configuration values */
	init_gui();
	init_iir();
	on = eqcfg.equalizer_active;
	volume_diff=0;
}

static void cleanup(void)
{
	clean_gui();
}

void onswitch(gboolean onstatus)
{
	on = onstatus;
	if (eqcfg.auto_volume_down)
	{
	    if (!onstatus)
	        mainwin_set_volume_diff(-35);
	    else {
	        volume_diff=35;
		    (void)gtk_timeout_add(eqcfg.auto_volume_down_ms, raisevol, NULL);
	    }
	}
}

void set_eq_value(float value, gint index, gint chn)
{
	/* Map the gain and preamp values */
	if (index >= 0)
		gain[index][chn] = 0.03 * value + 
							0.000999999 * value * value;
	else
 		/* -12dB .. 12dB mapping */
 		preamp[chn] = 1.0 + 0.0932471 * value + 0.00279033 * value * value;
}

void set_iir_func()
{
	if (eqcfg.use_fixed)
		eq_ep.mod_samples = mod_samples_fixed;
	else
		eq_ep.mod_samples = mod_samples;
}


gint raisevol(gpointer data)
{
	mainwin_set_volume_diff(  1  );
	if (--volume_diff == 0) 
		return FALSE;
	return TRUE;
}

/* 
 * The main stuff
 */
int mod_samples(gpointer * d, gint length, AFormat afmt, gint srate, gint nch)
{
	/* Support all this later */
	if (!on || !(afmt == FMT_S16_NE ||
	      (afmt == FMT_S16_LE && G_BYTE_ORDER == G_LITTLE_ENDIAN) ||
	      (afmt == FMT_S16_BE && G_BYTE_ORDER == G_BIG_ENDIAN)) ||
		  srate != 44100)
		return length;
	
	return iir(d, length);
}

int mod_samples_fixed(gpointer * d, gint length, AFormat afmt, gint srate, gint nch)
{
	/* Support all this later */
	if (!on || !(afmt == FMT_S16_NE ||
	      (afmt == FMT_S16_LE && G_BYTE_ORDER == G_LITTLE_ENDIAN) ||
	      (afmt == FMT_S16_BE && G_BYTE_ORDER == G_BIG_ENDIAN)) ||
		  srate != 44100)
		return length;
	return iir_fixed(d, length);
}
