/*
 *   Copyright (C) 2002  Felipe Rivera <liebremx at users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 *   Calculate coefficients for the IIR filters of the equalizer
 *
 *   $Id: calc_bands.c,v 1.1.1.1 2003/02/24 01:58:30 liebremx Exp $
 */

#include <stdio.h>
#include <math.h>
#include <gsl/gsl_math.h>


/*
 *  Can't check this because of missing GSL lib
 */

/*  
 *  the following lines are intentially oversized to be able
 *  to check the consistency of the name and frequency arrays 
 */
 
/* 
 *  There are two strategies are possible to select middle frequencies.
 *  One is to round to ISO frequencies, one is to use equal spaced frequencies and not to care about
 *  non-integer numbers. What is the better one?
 */ 

#define EQU_SETTINGS	9

/* Original frequencies in the 10-band equalizer in XMMS */
const char*  band_names_original10 [] = { "60", "170", "310", "600", "1k", "3k", "6k", "12k", "14k", "16k" };
const double band_original_f010    [] = {  60 ,  170 ,  310 ,  600 ,  1e3,  3e3,  6e3,  12e3,  14e3,  16e3 };

/* 0.6 bands per octave equalizer */
const char*  band_names06          [] = { "32", "100", "320", "1k", "3k" , "10k" };
const double band_f006             [] = {  32 ,  100 ,  320 ,  1e3,  3200,  10e3 };

/* 0.75 bands per octave equalizer */
const char*  band_names08          [] = { "25", "63", "160", "400", "1k", "2.5k", "6.3k", "16k" };
const double band_f008             [] = {  25 ,  63 ,  160 ,  400 ,  1e3,  2500 ,  6300 ,  16e3 };

/* 1 band per octave equalizer */
const char*  band_names10          [] = { "32" , "63", "125", "250", "500", "1k", "2k", "4k", "8k", "16k" };
const double band_f010             [] = { 31.25, 62.5,  125 ,  250 ,  500 ,  1e3,  2e3,  4e3,  8e3,  16e3 };

/* 1.5 bands per octave equalizer */
const char*  band_names15          [] = { "25", "40", "63", "100", "160", "250", "400", "630", "1k", "1.6k", "2.5k", "4k", "6.3k", "10k", "16k" };
const double band_f015             [] = {  25  , 40  , 63  , 100  , 160 ,  250 ,  400 ,  630,   1e3,  1600 ,  2500 ,  4e3,  6300  , 10e3,  16e3 };

/* 2 bands per octave equalizer */
const char*  band_names21          [] = { "22", "32", "45", "63", "90", "125", "175", "250", "350", "500", "700", "1k", "1.4k", "2k", "2.8k", "4k", "5.6k", "8k", "11k", "16k", "22k" };
const double band_f021             [] = {  22 ,  32 ,  44 ,  63 ,  90 ,  125 ,  175 ,  250 ,  350 ,  500 ,  700 ,  1e3,  1400 ,  2e3,  2800 ,  4e3,  5600 ,  8e3,  11e3,  16e3,  22e3 };

/* 2 bands per octave equalizer */
const char*  band_names25          [] = { "25", "31.5", "40", "50", "80", "100", "125", "160", "250", "315", "400", "500", "800", "1k", "1.25k", "1.6k", "2.5k", "3.15k", "4k", "5k", "8k", "10k", "12.5k", "16k", "20k" };
const double  band_f025          [] = { 25, 31.5, 40, 50, 80, 100, 125, 160, 250, 315, 400, 500, 800, 1e3, 1.25e3, 1.6e3, 2.5e3, 3.15e3, 4e3, 5e3, 8e3, 10e3, 12.5e3, 16e3, 20e3 };

/* 3 bands per octave equalizer */
const char*  band_names31          [] = { "20", "25", "32" , "40", "50", "63", "80", "100", "125", "160", "200", "250", "315", "400", "500", "630", "800", "1k", "1.25k", "1.6k", "2k", "2.5k", "3.2k", "4k", "5k", "6.3k", "8k", "10k", "12.5k", "16k", "20k" };
const double band_f031             [] = {  20,   25,   31.5,  40 ,  50 ,  63 ,  80 ,  100 ,  125 ,  160 ,  200 ,  250 ,  315 ,  400 ,  500 ,  630 ,  800 , 1e3 ,  1250  ,  1600 ,  2e3,  2500 ,  3150 ,  4e3,  5e3,  6300 ,  8e3,  10e3,  12500 ,  16e3,  20e3 };

/* 6 bands per octave equalizer */
const char*  band_names61          [] = { "20", "22.5", "25", "28", "31.5", "35.5", "40", "45", "50", "56", "63", "71", "80", "90", "100", "112", "125", "140", "160", "180", "200", "225", "250", "280", "315", "355", "400", "450", "500", "560", "630", "710", "800", "900", "1k", "1120", "1250", "1400", "1600", "1800", "2k", "2250", "2500", "2800", "3150", "3550", "4k", "4500", "5k", "5600", "6300", "7100", "8k", "9k", "10k", "11200", "12500", "14k", "16k", "18k", "20k" };
const double band_f061             [] = {  20 ,  22.5 ,  25 ,  28 ,  31.5 ,  35.5 ,  40 ,  45 ,  50 ,  56 ,  63 ,  71 ,  80 ,  90 ,  100 ,  112 ,  125 ,  140 ,  160 ,  180 ,  200 ,  225 ,  250 ,  280 ,  315 ,  355 ,  400 ,  450 ,  500 ,  560 ,  630 ,  710 ,  800 ,  900 ,  1e3,  1120 ,  1250 ,  1400 ,  1600 ,  1800 ,  2e3,  2250 ,  2500 ,  2800 ,  3150 ,  3550 ,  4e3,  4500 ,  5e3,  5600 ,  6300 ,  7100 ,  8e3,  9e3,  10e3,  11200 ,  12500 ,  14e3,  16e3,  18e3,  20e3 };


const double*  cfs              [EQU_SETTINGS] = { 
    band_original_f010, 
    band_f006, 
    band_f008, 
    band_f010, 
    band_f015, 
    band_f021, 
    band_f025, 
    band_f031,
    band_f061,
};

double octave[EQU_SETTINGS] = { 
    1.0, 
    1.0, 
    1.0,
    1.0,
    2.0/3.0, 
    1.0/3.0, 
    1.0/3.0,
    1.0/3.0, 
    1.0/6.0 
};

const char**   band_names       [EQU_SETTINGS] = { 
    band_names_original10, 
    band_names06, 
    band_names08, 
    band_names10, 
    band_names15, 
    band_names21, 
    band_names25, 
    band_names31,
    band_names61,
};
const int      band_count       [EQU_SETTINGS] = { 
    sizeof(band_original_f010)/sizeof(*band_f010), 
    sizeof(band_f006)         /sizeof(*band_f010), 
    sizeof(band_f008)         /sizeof(*band_f010), 
    sizeof(band_f010)         /sizeof(*band_f010), 
    sizeof(band_f015)         /sizeof(*band_f010), 
    sizeof(band_f021)         /sizeof(*band_f010), 
    sizeof(band_f025)         /sizeof(*band_f010), 
    sizeof(band_f031)         /sizeof(*band_f010), 
    sizeof(band_f061)         /sizeof(*band_f010), 
};
const char*    band_struct_name [EQU_SETTINGS] = { 
    "original10", 
    "06", 
    "08", 
    "10", 
    "15", 
    "21",
    "25",
    "31",
    "61",
};


#define GAIN_F0 1.0
#define GAIN_F1 GAIN_F0 / M_SQRT2

#define SAMPLING_FREQ 44100.0
#define TETA(f) (2*M_PI*(double)f/SAMPLING_FREQ)
#define TWOPOWER(value) (value * value)

#define BETA2(tf0, tf) \
                (TWOPOWER(GAIN_F1)*TWOPOWER(cos(tf0)) \
                 - 2.0 * TWOPOWER(GAIN_F1) * cos(tf) * cos(tf0) \
                 + TWOPOWER(GAIN_F1) \
                - TWOPOWER(GAIN_F0) * TWOPOWER(sin(tf)))
#define BETA1(tf0, tf) \
                (2.0 * TWOPOWER(GAIN_F1) * TWOPOWER(cos(tf)) \
                 + TWOPOWER(GAIN_F1) * TWOPOWER(cos(tf0)) \
                - 2.0 * TWOPOWER(GAIN_F1) * cos(tf) * cos(tf0) \
                - TWOPOWER(GAIN_F1) + TWOPOWER(GAIN_F0) * TWOPOWER(sin(tf)))
#define BETA0(tf0, tf) \
                (0.25 * TWOPOWER(GAIN_F1) * TWOPOWER(cos(tf0)) \
                 - 0.5 * TWOPOWER(GAIN_F1) * cos(tf) * cos(tf0) \
                + 0.25 * TWOPOWER(GAIN_F1) \
                - 0.25 * TWOPOWER(GAIN_F0) * TWOPOWER(sin(tf)))

#define GAMMA(beta, tf0) ((0.5 + beta) * cos(tf0))
#define ALPHA(beta) ((0.5 - beta)/2.0)


/*
 *  Get the lower and the upper limit at both sides of the middle frequency fmid. 
 *  These will be cut at -3 dB.
 *  These will be calculated by the geometric mean of the two neighbour equalizer sliders.
 *  For the first and last slider the distance will be continued.
 */

/* Get the freqs at both sides of F0. These will be cut at -3dB */
void find_f1_and_f2(double f0, double octave_percent,
					double *f1, double *f2)
{
	double octave_factor = pow(2.0, octave_percent/2.0);
	*f1 = ceilf(f0/octave_factor);
	*f2 = ceilf(f0*octave_factor);
}

static void 
ffind_f1_and_f2 ( double  fmid , 
                 double  fprev,          /* 0. for first band */
		 double  fnext,		 /* 0. for last  band */
                 double* flow , 
		 double* fhigh )
{
    if ( fprev != 0. )
        *flow  = sqrt ( fprev * fmid         );
    else
        *flow  = sqrt (  fmid * fmid / fnext );
	
    if ( fnext != 0. )
        *fhigh = sqrt (         fmid * fnext );
    else
        *fhigh = sqrt (  fmid * fmid / fprev );
}


int main(int argc, char *argv[])
{
    FILE *file, *file_fixed;
    int i, n, roots;
    double f1, f2;
    double x0, x1;

    file = fopen("iir_cf.h", "w+");
    file_fixed = fopen("iir_fixed_cf.h", "w+");
    fprintf(file, "\n/* BETA, ALPHA, GAMMA */\n");

  for ( n = 0; n < EQU_SETTINGS; n++ )
  {
    double *coeff = (double *)cfs[n];
    char **names = (char **)band_names[n];
    
    fprintf ( file, 
              "static sIIRCoefficients iir_cf%s[] __attribute__((aligned)) = {\n",
              band_struct_name[n] );
    fprintf ( file_fixed, 
              "static sIIRCoefficients_fixed iir_cf%s_fixed[] __attribute__((aligned)) = {\n",
              band_struct_name[n] );
    
    for (i=0; i<band_count[n]; i++)
    {
        double alpha, beta, gamma;

        /* Find -3dB frequencies for the center freq */
        ffind_f1_and_f2 ( coeff[i], 
	                 i > 0                ?  coeff[i-1]  :  0., 
			 i < band_count[n]-1  ?  coeff[i+1]  :  0., 
			 &f1, &f2 );

       	find_f1_and_f2(coeff[i], octave[n], &f1, &f2);

        printf ( "Band %-5.5s: [%9.3f, %9.3f, %9.3f]\n", names [i], f1, coeff[i], f2 );
        /* Find Beta */
        roots = gsl_poly_solve_quadratic (
                    BETA2(TETA(coeff[i]), TETA(f1)),
                    BETA1(TETA(coeff[i]), TETA(f1)),
                    BETA0(TETA(coeff[i]), TETA(f1)),
                    &x0, &x1);
        if (roots > 0)
        {
            /* Got a solution, now calculate the rest of the factors */
            /* Take the smallest root always (which is the first returned
             * by gsl_poly_solve_quadratic
             *
             * NOTE: The IIR equation is
             *  y[n] = 2 * (alpha*(x[n]-x[n-2]) + gamma*y[n-1] - beta*y[n-2])
             *  Now the 2 factor has been distributed in the coefficients
             */
            x0 = x0;
            beta = 2.0 * x0;
            alpha = 2.0 * ALPHA(x0);
            gamma = 2.0 * GAMMA(x0, TETA(coeff[i]));

            /* Now write the coefficients */
            fprintf ( file, "    { (%.10e), (%.10e), (%.10e) },",
                            beta, alpha, gamma);
            fprintf ( file, "    /* %7.1f Hz */\n", coeff[i] );
            fprintf ( file_fixed, "    { EQ_REAL(%.10e), EQ_REAL(%.10e), EQ_REAL(%.10e) },",
                            beta, alpha, gamma);
            fprintf ( file_fixed, "    /* %7.1f Hz */\n", coeff[i] );
        }
        else
            printf("  **** Where are the roots?\n");
    }
    fprintf(file, "};\n");
    fprintf(file_fixed, "};\n");
  }
  fclose(file);
}
